-- vesa20.e
-- VESA 2.0 compatible routines for setting up SVGA graphics modes
-- Pete Eberlein <xseal@harborside.com>
-- April 22, 1999

---- FUNCTIONS --------------------------------------------------------------
--
-- vesa_mode(width, height, bits)
--  Returns 0 if the requested mode was set successfully.
--  Also sets the global variables listed below.
--  Bits can take 8,15,16,24,32.

global constant
    LINEAR = 1,     -- indicator for linear mode addressing
    BANKED = 2      -- indicator for banked addressing

global integer
    mode_type,  -- will be set to one of the two constants above

    bits_per_pixel,  -- will be set to one of 8,15,16,24,32

    bytes_per_pixel, -- the number of bytes per pixel

    bytes_per_line   -- the actual width of the screen in display memory
                     -- note: this may not be the same as the screen width
global atom
    video_mem        -- the address of display memory.  may be #A0000 or not
                     -- depending on whether the mode is banked or linear.
global integer
    video_mem_amount, -- the amount of RAM on your video card, in kilobytes

    video_pages      -- the max allowed number of pages for the current mode

---- PROCEDURES -------------------------------------------------------------
--
-- single_copy(atom kind, sequence arguments)
--  Does a fast copy of type kind (listed below) with the arguments given.
--  Arguments must be the correct length for kind (see below.)
--
-- list_copy(atom list_address)
--  Performs multiple copies using an allocated list of kind,followed by
--  the arguments,kind,arguments,etc. terminated by a 0.
--  The list must be allocated with 4 times the length of the list sequence.
--  poke4 should be used to poke the list sequence into the allocated memory.
--  ex: poke4(my_list, {kind1,arg1,arg2,...,kind2,arg1,arg2,...,0})
--  Calling conventions for custom routines are at the end of this file.

global atom
    fast_copy, -- This kind of copy is like to mem_copy, copies every byte.
               -- The arguments are: {source address, destination address,
               --   width, height, source width, destination width}

    clear_copy,-- This kind leaves out clear pixels, so that the background
               -- shows through... good for sprites, etc.
               -- The arguments are: {source address, destination address,
               --   width, height, source width, destination width}

    bank_copy, -- This kind copies pixels the same as fast_copy, only it
               -- performs bank-switching when necessary.
               -- The arguments are: {source address, destination address,
               --   width, height, source width, destination width}

    mix_copy,  -- This kind translates pixels using a lookup table based on
               -- the source and destination pixels.  The table must be
               -- 65536 bytes of allocated memory.
               -- The offset calculation into the lookup table is:
               --   dest = peek(table_address + source_pixel + 256*dest_pixel)
               -- The arguments are: {source address, destination address,
               --   width, height, source width, destination width,
               --   table address}

    map_copy,  -- This kind translates pixels using a lookup table of
               -- 2-byte relative addresses.  When a pixel is drawn, the
               -- value is loaded from the lookup table.  Then that value
               -- is added to the source address and the pixel at that
               -- location is copied to the destination pixel.
               -- The arguments are: {table address, destination address,
               --   width, height, table width, destination width,
               --   source address}

    screen_copy -- This will be set, after a call to vesa_mode, to either
                -- fast_copy or bank_copy, depending on whether the mode
                -- is linear or banked.
                -- Use screen_copy whenever you copy to video_mem.


-----------------------------------------------------------------------------


include machine.e

-----------------------------------------------------------------
-- VESA (SuperVGA) modes ----------------------------------------
-----------------------------------------------------------------

constant bank_proc = allocate(17)
constant bank_regs = allocate(#32)


-- something in here really pisses off NT, so lets say hello and play nice
constant detectNT = equal(getenv("OS"), "Windows_NT")


integer bank_granularity, bank_start
atom win_func_ptr

-- standard modes, only used if there is no mode list in rom
constant VESAModes = {
    {640,400,0,0,0,8,0,0,0}, -- 256
    {640,480,0,0,0,8,0,0,0}, -- 257
    {800,600,0,0,0,4,0,0,0}, -- 258
    {800,600,0,0,0,8,0,0,0}, -- 259
    {1024,768,0,0,0,4,0,0,0}, -- 260
    {1024,768,0,0,0,8,0,0,0}, -- 261
    {1280,1024,0,0,0,4,0,0,0}, -- 262
    {1280,1024,0,0,0,8,0,0,0}, -- 263
    {}, -- 264
    {}, -- 265
    {}, -- 266
    {}, -- 267
    {}, -- 268
    {320,200,0,0,0,15,0,0,0}, -- 269
    {320,200,0,0,0,16,0,0,0}, -- 270
    {320,200,0,0,0,24,0,0,0}, -- 271
    {640,480,0,0,0,15,0,0,0}, -- 272
    {640,480,0,0,0,16,0,0,0}, -- 273
    {640,480,0,0,0,24,0,0,0}, -- 274
    {800,600,0,0,0,15,0,0,0}, -- 275
    {800,600,0,0,0,16,0,0,0}, -- 276
    {800,600,0,0,0,24,0,0,0}, -- 277
    {1024,768,0,0,0,15,0,0,0}, -- 278
    {1024,768,0,0,0,16,0,0,0}, -- 279
    {1024,768,0,0,0,24,0,0,0}, -- 280
    {1280,1024,0,0,0,15,0,0,0}, -- 281
    {1280,1024,0,0,0,16,0,0,0}, -- 282
    {1280,1024,0,0,0,24,0,0,0}  -- 283
    }

-- code to convert the physical base address to the linear frame buffer address
constant getlfb = allocate(68)
poke(getlfb,{
    #60,                    --    0: pusha
    #8B,#1D,#D4,#6A,#CC,#82,--    1: mov ebx, [@linear]
    #85,#DB,                --    7: test ebx, ebx
    #74,#0C,                --    9: jz batman
    #89,#D9,                --    B: mov ecx, ebx
    #C1,#EB,#10,            --    D: shr ebx, 16
    #B8,#01,#08,#00,#00,    --   10: mov eax, #801
    #CD,#31,                --   15: int #31
                            --   17: batman: 
    #BB,#00,#00,#00,#00,    --   17: mov ebx, PhysBasePtr (24)
    #89,#D9,                --   1C: mov ecx, ebx
    #C1,#EB,#10,            --   1E: shr ebx, 16
    #BE,#00,#00,#40,#00,    --   21: mov esi, 4194304
    #89,#F7,                --   26: mov edi, esi
    #C1,#EE,#10,            --   28: shr esi, 16
    #B8,#00,#08,#00,#00,    --   2B: mov eax, #800
    #CD,#31,                --   30: int #31
    #C1,#E3,#10,            --   32: shl ebx, 16
    #66,#89,#CB,            --   35: mov bx, cx
    #89,#1D,#D4,#6A,#CC,#82,--   38: mov [@linear], ebx
    #61,                    --   3E: popa
    #C3,                    --   3F: ret
    #00,#00,#00,#00})       --   40: linear: dd 0 (64)
poke4(getlfb + 3, getlfb + 64) -- @linear
poke4(getlfb + 58, getlfb + 64) -- @linear


global function vesa_mode(integer width, integer height, integer bpp)
--> attempts to set a VESA mode
--> width,height pairs may be (640,480),(800,600),(1024,768),(1280,1024)
-->   (other modes may be supported by your video hardware)
--> bpp may be 8, 15, 16, 24 or 32
-->   (some video cards support only 15 or 16, and 24 or 32, not both)
--> returns 0 on success, 1 otherwise
--> See also: force_vesa, print_vesa
    atom data, OEMStringPtr, VideoModePtr, ModeAttributes
    atom PhysBasePtr, OffScreenMemOffset, OffScreenMemSize
    integer bank_order, bank_64k
    sequence regs
    object result, VESASignature, VESAVersion, OEMString, Capabilities, VideoModes
    regs = repeat(0,10)
    regs[REG_AX] = #4F00
    data = allocate_low(256)
    regs[REG_ES] = floor(data / 16)
    regs[REG_DI] = and_bits(data, 15)
    regs = dos_interrupt(#10, regs)
    if regs[REG_AX] != #4F then
        return 1
    end if
    
    VESASignature = peek({data,4})
    VESAVersion = 0.01*peek(data+4)+peek(data+5)
    OEMStringPtr = peek(data+6) + 256*peek(data+7) + 16*peek(data+8) + 4096*peek(data+9)
    Capabilities = peek4u(data+10)
    VideoModePtr = peek(data+14) + 256*peek(data+15) + 16*peek(data+16) + 4096*peek(data+17)
    free_low(data)
    OEMString = ""
    while peek(OEMStringPtr) do
        OEMString = append(OEMString, peek(OEMStringPtr))
        OEMStringPtr = OEMStringPtr + 1
    end while
    VideoModes = {}
    while compare(peek({VideoModePtr,2}),{#FF,#FF}) do
        VideoModes = append(VideoModes, peek(VideoModePtr)+256*peek(VideoModePtr+1))
        VideoModePtr = VideoModePtr + 2
    end while


    data = allocate_low(256)
    result = {}
    for mode = 1 to length(VideoModes) do
        regs = repeat(0,10)
        regs[REG_AX] = #4F01
        regs[REG_CX] = VideoModes[mode]
        regs[REG_ES] = floor(data / 16)
        regs[REG_DI] = and_bits(data, 15)
        regs = dos_interrupt(#10, regs)
        if regs[REG_AX] != #4F then
            free_low(data)
            return 1
        end if
        ModeAttributes = peek(data) + 256*peek(data+1)
        if 1-and_bits(ModeAttributes, 1) then
            result = {}  -- mode available but not supported, because
                         -- insufficient vram or incompatible monitor
        elsif and_bits(ModeAttributes,2) then
            result = (peek(data+18) + 256*peek(data+19)) &
                (peek(data+20) + 256*peek(data+21)) & peek({data+22,7})
        elsif VideoModes[mode] >= 256 and VideoModes[mode] - 255 <= length(VESAModes) then
            result = VESAModes[VideoModes[mode]-255]
        else
            result = {}  -- not enough information
        end if
        if length(result) then
--            if not ((bpp = 15 or bpp = 16) and (result[6] = 15 or result[6] = 16)) then
--                -- not a matching 15 or 16 bit mode
--            elsif not ((bpp = 24 or bpp = 32) and (result[6] = 24 or result[6] = 32)) then
--                -- not a matching 24 or 32 bit mode
            if not (bpp = result[6]) then
                -- not a matching other mode
            elsif width = result[1] and height = result[2] then

                bank_granularity = (peek(data + 4) + 256 * peek(data + 5)) * 1024
                bytes_per_line = peek(data + #10) + 256 * peek(data + #11)
                bits_per_pixel = result[6]
                bytes_per_pixel = floor((result[6]+7)/8)
                win_func_ptr = peek4u(data+12)
                --video_mem_amount = 

                regs[REG_AX] = #4F02
                regs[REG_BX] = VideoModes[mode]

                if and_bits(ModeAttributes, #80) and not detectNT then
                    -- linear frame buffer
                    regs[REG_BX] = regs[REG_BX] + #4000
                    PhysBasePtr = peek4u(data+#28)
                    OffScreenMemOffset = peek4u(data+#2C)
                    OffScreenMemSize = peek(data+#30) + 256 * peek(data+#31)

                    mode_type = LINEAR
                    screen_copy = fast_copy

                else
                    -- banked
                    mode_type = BANKED
                    screen_copy = bank_copy
                end if
                
                regs = dos_interrupt(#10, regs)
                result = regs[REG_AX]  -- success

                exit
            end if
        end if
    end for
    free_low(data)
    if atom(result) then

        if mode_type = LINEAR then

            poke4(getlfb + 24, PhysBasePtr)  -- convert physical base address
            call(getlfb)                     -- to linear frame buffer
            video_mem = peek4u(getlfb + 64)

        else

            video_mem = #A0000

            poke4(bank_regs + #2A, win_func_ptr) -- CS:IP
            poke4(bank_regs + #10, 0) -- BX
            poke4(bank_regs + #14, 1) -- DX
            poke(bank_proc,
               {#60,                    -- pusha
                #B8,#02,#FF,#00,#00,    -- mov eax, #FF02
                #1E,                    -- push ds
                #07,                    -- pop es
                #BF}&int_to_bytes(bank_regs)&{-- mov edi, bank_regs
                #CD,#31,                -- int #31
                #61,                    -- popa
                #C3})                   -- ret
            bank_start = 0

            bank_64k = #10000 / bank_granularity
            bank_order = 1
            while power(2, bank_order) < bank_granularity do
                bank_order = bank_order + 1            
            end while

            poke(bank_copy + 72, bank_order)
            poke4(bank_copy + 85, bank_regs)
            poke(bank_copy + 94, bank_order)
            poke4(bank_copy + 210, bank_64k)
            poke4(bank_copy + 239, bank_regs)

        end if

    end if
    return sequence(result)
end function


----------------------------------------------------------------------------
atom start_copy
constant single_list = allocate(64)

global procedure single_copy(atom kind, sequence parameters)
    poke4(single_list, kind & parameters & 0)
    poke4(start_copy + 2, single_list)
    call(start_copy)
end procedure

global procedure list_copy(atom list_address)
    poke4(start_copy + 2, list_address)
    call(start_copy)
end procedure

----------------------------------------------------------------------------

start_copy = allocate(16)
poke(start_copy, {
    #60,                    --    0: pusha
    #BB,#00,#00,#00,#00,    --    1: mov ebx, list_address (2)
    #8B,#03,                --    6: mov eax, [ebx]
    #85,#C0,                --    8: test eax, eax
    #74,#02,                --    A: jz done
    #FF,#E0,                --    C: jmp near eax
    #61,                    --    E: done: popa
    #C3})                   --    F: ret

fast_copy = allocate(159)
poke(fast_copy, {
    #8B,#73,#04,            --    0: mov esi, [ebx+4]
    #8B,#7B,#08,            --    3: mov edi, [ebx+8]
    #8B,#4B,#0C,            --    6: mov ecx, [ebx+12]
    #8B,#6B,#10,            --    9: mov ebp, [ebx+16]
    #8B,#43,#14,            --    C: mov eax, [ebx+20]
    #29,#C8,                --    F: sub eax, ecx
    #A3,#EB,#6B,#C5,#83,    --   11: mov [@source_add], eax
    #8B,#43,#18,            --   16: mov eax, [ebx+24]
    #83,#C3,#1C,            --   19: add ebx, 28
    #29,#C8,                --   1C: sub eax, ecx
    #A3,#F1,#6B,#C5,#83,    --   1E: mov [@dest_add], eax
    #89,#1D,#F9,#6B,#C5,#83,--   23: mov [@saved_ebx], ebx
    #89,#FB,                --   29: mov ebx, edi
    #75,#13,                --   2B: jnz noopt
    #0B,#05,#EB,#6B,#C5,#83,--   2D: or eax, [@source_add]
    #75,#0B,                --   33: jnz noopt
    #89,#C8,                --   35: mov eax, ecx
    #F7,#E5,                --   37: mul ebp
    #BD,#01,#00,#00,#00,    --   39: mov ebp, 1
    #89,#C1,                --   3E: mov ecx, eax
    #F7,#DB,                --   40: noopt: neg ebx
    #81,#E3,#03,#00,#00,#00,--   42: and ebx, #100000003
    #29,#D9,                --   48: sub ecx, ebx
    #7D,#04,                --   4A: jge line1ok
    #01,#CB,                --   4C: add ebx, ecx
    #31,#C9,                --   4E: xor ecx, ecx
    #89,#1D,#D5,#6B,#C5,#83,--   50: line1ok: mov [@line1], ebx
    #88,#CB,                --   56: mov bl, cl
    #C1,#E9,#02,            --   58: shr ecx, 2
    #80,#E3,#03,            --   5B: and bl, 3
    #89,#0D,#DC,#6B,#C5,#83,--   5E: mov [@line2], ecx
    #89,#1D,#E3,#6B,#C5,#83,--   64: mov [@line3], ebx
    #EB,#00,                --   6A: jmp line_loop
    #B9,#00,#00,#00,#00,    --   6C: line_loop: mov ecx, line1 (109)
    #F3,#A4,                --   71: line_start: rep movsb
    #B9,#00,#00,#00,#00,    --   73: mov ecx, line2 (116)
    #F3,#A5,                --   78: rep movsd
    #B9,#00,#00,#00,#00,    --   7A: mov ecx, line3 (123)
    #F3,#A4,                --   7F: rep movsb
    #81,#C6,#00,#00,#00,#00,--   81: add esi, source_add (131)
    #81,#C7,#00,#00,#00,#00,--   87: add edi, dest_add (137)
    #4D,                    --   8D: dec ebp
    #75,#DC,                --   8E: jnz line_loop
    #BB,#00,#00,#00,#00,    --   90: mov ebx, saved_ebx (145)
    #8B,#03,                --   95: mov eax, [ebx]
    #85,#C0,                --   97: test eax, eax
    #74,#02,                --   99: jz done
    #FF,#E0,                --   9B: jmp near eax
    #61,                    --   9D: done: popa
    #C3})                   --   9E: ret
poke4(fast_copy + 18, fast_copy + 131) -- @source_add
poke4(fast_copy + 31, fast_copy + 137) -- @dest_add
poke4(fast_copy + 37, fast_copy + 145) -- @saved_ebx
poke4(fast_copy + 47, fast_copy + 131) -- @source_add
poke4(fast_copy + 82, fast_copy + 109) -- @line1
poke4(fast_copy + 96, fast_copy + 116) -- @line2
poke4(fast_copy + 102, fast_copy + 123) -- @line3

clear_copy = allocate(268)
poke(clear_copy, {
    #8B,#73,#04,            --    0: mov esi, [ebx+4]
    #8B,#7B,#08,            --    3: mov edi, [ebx+8]
    #8B,#4B,#0C,            --    6: mov ecx, [ebx+12]
    #8B,#6B,#10,            --    9: mov ebp, [ebx+16]
    #8B,#43,#14,            --    C: mov eax, [ebx+20]
    #29,#C8,                --    F: sub eax, ecx
    #A3,#75,#9B,#C5,#83,    --   11: mov [@source_add], eax
    #8B,#43,#18,            --   16: mov eax, [ebx+24]
    #83,#C3,#1C,            --   19: add ebx, 28
    #29,#C8,                --   1C: sub eax, ecx
    #A3,#7B,#9B,#C5,#83,    --   1E: mov [@dest_add], eax
    #89,#1D,#83,#9B,#C5,#83,--   23: mov [@saved_ebx], ebx
    #89,#FB,                --   29: mov ebx, edi
    #F7,#DB,                --   2B: neg ebx
    #81,#E3,#03,#00,#00,#00,--   2D: and ebx, #100000003
    #29,#D9,                --   33: sub ecx, ebx
    #7D,#04,                --   35: jge line1ok
    #01,#CB,                --   37: add ebx, ecx
    #31,#C9,                --   39: xor ecx, ecx
    #89,#1D,#56,#9B,#C5,#83,--   3B: line1ok: mov [@line1], ebx
    #88,#CB,                --   41: mov bl, cl
    #80,#E3,#03,            --   43: and bl, 3
    #C1,#E9,#02,            --   46: shr ecx, 2
    #75,#08,                --   49: jnz line2ok
    #01,#1D,#56,#9B,#C5,#83,--   4B: add dword ptr [@line1], ebx
    #89,#CB,                --   51: mov ebx, ecx
    #89,#0D,#6D,#9B,#C5,#83,--   53: line2ok: mov [@line2], ecx
    #89,#1D,#D3,#9B,#C5,#83,--   59: mov [@line3], ebx
    #EB,#00,                --   5F: jmp line_loop
    #B9,#00,#00,#00,#00,    --   61: line_loop: mov ecx, line1 (98)
    #85,#C9,                --   66: test ecx, ecx
    #74,#0D,                --   68: jz end_loop1
    #8A,#06,                --   6A: pixel_loop1: mov al, [esi]
    #84,#C0,                --   6C: test al, al
    #74,#02,                --   6E: jz clear1
    #88,#07,                --   70: mov [edi], al
    #46,                    --   72: clear1: inc esi
    #47,                    --   73: inc edi
    #49,                    --   74: dec ecx
    #75,#F3,                --   75: jnz pixel_loop1
    #81,#C1,#00,#00,#00,#00,--   77: end_loop1: add ecx, line2 (121)
    #75,#1E,                --   7D: jnz pixel_loop2
    #81,#C6,#00,#00,#00,#00,--   7F: end_loop2: add esi, source_add (129)
    #81,#C7,#00,#00,#00,#00,--   85: add edi, dest_add (135)
    #4D,                    --   8B: dec ebp
    #75,#D3,                --   8C: jnz line_loop
    #BB,#00,#00,#00,#00,    --   8E: exit_loop: mov ebx, saved_ebx (143)
    #8B,#03,                --   93: mov eax, [ebx]
    #85,#C0,                --   95: test eax, eax
    #74,#02,                --   97: jz done
    #FF,#E0,                --   99: jmp near eax
    #61,                    --   9B: done: popa
    #C3,                    --   9C: ret
    #8B,#06,                --   9D: pixel_loop2: mov eax, [esi]
    #85,#C0,                --   9F: test eax, eax
    #74,#31,                --   A1: jz clear2e
    #89,#C3,                --   A3: mov ebx, eax
    #C1,#EB,#10,            --   A5: shr ebx, 16
    #84,#C0,                --   A8: test al, al
    #74,#10,                --   AA: jz clear2a
    #84,#E4,                --   AC: test ah, ah
    #74,#42,                --   AE: jz clear2x
    #84,#DB,                --   B0: test bl, bl
    #74,#43,                --   B2: jz clear2y
    #84,#FF,                --   B4: test bh, bh
    #74,#47,                --   B6: jz clear2z
    #89,#07,                --   B8: mov [edi], eax
    #EB,#18,                --   BA: jmp clear2e
    #47,                    --   BC: clear2a: inc edi
    #84,#E4,                --   BD: test ah, ah
    #74,#02,                --   BF: jz clear2b
    #88,#27,                --   C1: mov [edi], ah
    #47,                    --   C3: clear2b: inc edi
    #84,#DB,                --   C4: test bl, bl
    #74,#02,                --   C6: jz clear2c
    #88,#1F,                --   C8: mov [edi], bl
    #47,                    --   CA: clear2c: inc edi
    #84,#FF,                --   CB: test bh, bh
    #74,#02,                --   CD: jz clear2d
    #88,#3F,                --   CF: mov [edi], bh
    #83,#EF,#03,            --   D1: clear2d: sub edi, 3
    #83,#C7,#04,            --   D4: clear2e: add edi, 4
    #83,#C6,#04,            --   D7: add esi, 4
    #49,                    --   DA: dec ecx
    #75,#C0,                --   DB: jnz pixel_loop2
    #81,#C1,#00,#00,#00,#00,--   DD: add ecx, line3 (223)
    #74,#9A,                --   E3: pixel_loop3: jz end_loop2
    #8A,#06,                --   E5: mov al, [esi]
    #84,#C0,                --   E7: test al, al
    #74,#02,                --   E9: jz clear3
    #88,#07,                --   EB: mov [edi], al
    #46,                    --   ED: clear3: inc esi
    #47,                    --   EE: inc edi
    #49,                    --   EF: dec ecx
    #EB,#F1,                --   F0: jmp pixel_loop3
    #88,#07,                --   F2: clear2x: mov [edi], al
    #47,                    --   F4: inc edi
    #EB,#CC,                --   F5: jmp clear2b
    #88,#07,                --   F7: clear2y: mov [edi], al
    #47,                    --   F9: inc edi
    #88,#27,                --   FA: mov [edi], ah
    #47,                    --   FC: inc edi
    #EB,#CB,                --   FD: jmp clear2c
    #88,#07,                --   FF: clear2z: mov [edi], al
    #47,                    --  101: inc edi
    #88,#27,                --  102: mov [edi], ah
    #47,                    --  104: inc edi
    #88,#1F,                --  105: mov [edi], bl
    #83,#EF,#02,            --  107: sub edi, 2
    #EB,#C8})               --  10A: jmp clear2e
poke4(clear_copy + 18, clear_copy + 129) -- @source_add
poke4(clear_copy + 31, clear_copy + 135) -- @dest_add
poke4(clear_copy + 37, clear_copy + 143) -- @saved_ebx
poke4(clear_copy + 61, clear_copy + 98) -- @line1
poke4(clear_copy + 77, clear_copy + 98) -- @line1
poke4(clear_copy + 85, clear_copy + 121) -- @line2
poke4(clear_copy + 91, clear_copy + 223) -- @line3

mix_copy = allocate(338)
poke(mix_copy, {
    #8B,#73,#04,            --    0: mov esi, [ebx+4]
    #8B,#7B,#08,            --    3: mov edi, [ebx+8]
    #8B,#4B,#0C,            --    6: mov ecx, [ebx+12]
    #8B,#6B,#10,            --    9: mov ebp, [ebx+16]
    #8B,#43,#14,            --    C: mov eax, [ebx+20]
    #29,#C8,                --    F: sub eax, ecx
    #A3,#74,#05,#C6,#83,    --   11: mov [@source_add], eax
    #8B,#43,#18,            --   16: mov eax, [ebx+24]
    #29,#C8,                --   19: sub eax, ecx
    #A3,#7A,#05,#C6,#83,    --   1B: mov [@dest_add], eax
    #8B,#43,#1C,            --   20: mov eax, [ebx+28]
    #A3,#5F,#05,#C6,#83,    --   23: mov [@table1], eax
    #83,#C3,#20,            --   28: add ebx, 32
    #A3,#CE,#05,#C6,#83,    --   2B: mov [@table2], eax
    #89,#1D,#82,#05,#C6,#83,--   30: mov [@saved_ebx], ebx
    #A3,#E1,#05,#C6,#83,    --   36: mov [@table3], eax
    #89,#FB,                --   3B: mov ebx, edi
    #A3,#F0,#05,#C6,#83,    --   3D: mov [@table4], eax
    #A3,#03,#06,#C6,#83,    --   42: mov [@table5], eax
    #A3,#0F,#06,#C6,#83,    --   47: mov [@table6], eax
    #F7,#DB,                --   4C: neg ebx
    #31,#C0,                --   4E: xor eax, eax
    #81,#E3,#03,#00,#00,#00,--   50: and ebx, #100000003
    #29,#D9,                --   56: sub ecx, ebx
    #7D,#04,                --   58: jge line1ok
    #01,#CB,                --   5A: add ebx, ecx
    #31,#C9,                --   5C: xor ecx, ecx
    #89,#1D,#4D,#05,#C6,#83,--   5E: line1ok: mov [@line1], ebx
    #88,#CB,                --   64: mov bl, cl
    #80,#E3,#03,            --   66: and bl, 3
    #C1,#E9,#02,            --   69: shr ecx, 2
    #75,#08,                --   6C: jnz line2ok
    #01,#1D,#4D,#05,#C6,#83,--   6E: add dword ptr [@line1], ebx
    #89,#CB,                --   74: mov ebx, ecx
    #89,#0D,#6C,#05,#C6,#83,--   76: line2ok: mov [@line2], ecx
    #89,#1D,#BE,#05,#C6,#83,--   7C: mov [@line3], ebx
    #EB,#00,                --   82: jmp line_loop
    #B9,#00,#00,#00,#00,    --   84: line_loop: mov ecx, line1 (133)
    #85,#C9,                --   89: test ecx, ecx
    #74,#15,                --   8B: jz end_loop1
    #8A,#06,                --   8D: pixel_loop1: mov al, [esi]
    #84,#C0,                --   8F: test al, al
    #74,#0A,                --   91: jz clear1
    #8A,#27,                --   93: mov ah, [edi]
    #8A,#80,#00,#00,#00,#00,--   95: mov al, [eax + table1] (151)
    #88,#07,                --   9B: mov [edi], al
    #46,                    --   9D: clear1: inc esi
    #47,                    --   9E: inc edi
    #49,                    --   9F: dec ecx
    #75,#EB,                --   A0: jnz pixel_loop1
    #81,#C1,#00,#00,#00,#00,--   A2: end_loop1: add ecx, line2 (164)
    #75,#1E,                --   A8: jnz pixel_loop2
    #81,#C6,#00,#00,#00,#00,--   AA: end_loop2: add esi, source_add (172)
    #81,#C7,#00,#00,#00,#00,--   B0: add edi, dest_add (178)
    #4D,                    --   B6: dec ebp
    #75,#CB,                --   B7: jnz line_loop
    #BB,#00,#00,#00,#00,    --   B9: exit_loop: mov ebx, saved_ebx (186)
    #8B,#03,                --   BE: mov eax, [ebx]
    #85,#C0,                --   C0: test eax, eax
    #74,#02,                --   C2: jz done
    #FF,#E0,                --   C4: jmp near eax
    #61,                    --   C6: done: popa
    #C3,                    --   C7: ret
    #8B,#16,                --   C8: pixel_loop2: mov edx, [esi]
    #85,#D2,                --   CA: test edx, edx
    #74,#1D,                --   CC: jz clear2e
    #8B,#1F,                --   CE: mov ebx, [edi]
    #84,#D2,                --   D0: test dl, dl
    #75,#3D,                --   D2: jnz clear2a
    #84,#F6,                --   D4: test dh, dh
    #75,#47,                --   D6: jnz clear2b
    #C1,#CA,#10,            --   D8: clear2x: ror edx, 16
    #84,#D2,                --   DB: test dl, dl
    #75,#51,                --   DD: jnz clear2c
    #C1,#CB,#10,            --   DF: clear2y: ror ebx, 16
    #84,#F6,                --   E2: test dh, dh
    #75,#5B,                --   E4: jnz clear2d
    #C1,#CB,#10,            --   E6: clear2z: ror ebx, 16
    #89,#1F,                --   E9: mov [edi], ebx
    #83,#C6,#04,            --   EB: clear2e: add esi, 4
    #83,#C7,#04,            --   EE: add edi, 4
    #49,                    --   F1: dec ecx
    #75,#D4,                --   F2: jnz pixel_loop2
    #81,#C1,#00,#00,#00,#00,--   F4: add ecx, dword line3 (246)
    #74,#AE,                --   FA: pixel_loop3: jz end_loop2
    #8A,#06,                --   FC: mov al, [esi]
    #84,#C0,                --   FE: test al, al
    #74,#0A,                --  100: jz clear3
    #8A,#27,                --  102: mov ah, [edi]
    #8A,#80,#00,#00,#00,#00,--  104: mov al, [eax + table2] (262)
    #88,#07,                --  10A: mov [edi], al
    #46,                    --  10C: clear3: inc esi
    #47,                    --  10D: inc edi
    #49,                    --  10E: dec ecx
    #EB,#E9,                --  10F: jmp pixel_loop3
    #88,#DC,                --  111: clear2a: mov ah, bl
    #88,#D0,                --  113: mov al, dl
    #84,#F6,                --  115: test dh, dh
    #8A,#98,#00,#00,#00,#00,--  117: mov bl, [eax + table3] (281)
    #74,#B9,                --  11D: jz clear2x
    #88,#FC,                --  11F: clear2b: mov ah, bh
    #88,#F0,                --  121: mov al, dh
    #C1,#CA,#10,            --  123: ror edx, 16
    #8A,#B8,#00,#00,#00,#00,--  126: mov bh, [eax + table4] (296)
    #84,#D2,                --  12C: test dl, dl
    #74,#AF,                --  12E: jz clear2y
    #C1,#CB,#10,            --  130: clear2c: ror ebx, 16
    #88,#D0,                --  133: mov al, dl
    #88,#DC,                --  135: mov ah, bl
    #84,#F6,                --  137: test dh, dh
    #8A,#98,#00,#00,#00,#00,--  139: mov bl, [eax + table5] (315)
    #74,#A5,                --  13F: jz clear2z
    #88,#F0,                --  141: clear2d: mov al, dh
    #88,#FC,                --  143: mov ah, bh
    #8A,#B8,#00,#00,#00,#00,--  145: mov bh, [eax + table6] (327)
    #C1,#CB,#10,            --  14B: ror ebx, 16
    #89,#1F,                --  14E: mov [edi], ebx
    #EB,#99})               --  150: jmp clear2e
poke4(mix_copy + 18, mix_copy + 172) -- @source_add
poke4(mix_copy + 28, mix_copy + 178) -- @dest_add
poke4(mix_copy + 36, mix_copy + 151) -- @table1
poke4(mix_copy + 44, mix_copy + 262) -- @table2
poke4(mix_copy + 50, mix_copy + 186) -- @saved_ebx
poke4(mix_copy + 55, mix_copy + 281) -- @table3
poke4(mix_copy + 62, mix_copy + 296) -- @table4
poke4(mix_copy + 67, mix_copy + 315) -- @table5
poke4(mix_copy + 72, mix_copy + 327) -- @table6
poke4(mix_copy + 96, mix_copy + 133) -- @line1
poke4(mix_copy + 112, mix_copy + 133) -- @line1
poke4(mix_copy + 120, mix_copy + 164) -- @line2
poke4(mix_copy + 126, mix_copy + 246) -- @line3

bank_copy = allocate(267)
poke(bank_copy, {
    #8B,#73,#04,            --    0: mov esi, [ebx+4]
    #8B,#7B,#08,            --    3: mov edi, [ebx+8]
    #8B,#4B,#0C,            --    6: mov ecx, [ebx+12]
    #8B,#6B,#10,            --    9: mov ebp, [ebx+16]
    #8B,#43,#14,            --    C: mov eax, [ebx+20]
    #29,#C8,                --    F: sub eax, ecx
    #A3,#4C,#11,#C6,#83,    --   11: mov [@source_add], eax
    #8B,#43,#18,            --   16: mov eax, [ebx+24]
    #83,#C3,#1C,            --   19: add ebx, 28
    #29,#C8,                --   1C: sub eax, ecx
    #A3,#52,#11,#C6,#83,    --   1E: mov [@dest_add], eax
    #89,#1D,#5A,#11,#C6,#83,--   23: mov [@saved_ebx], ebx
    #89,#FB,                --   29: mov ebx, edi
    #89,#0D,#75,#11,#C6,#83,--   2B: mov [@line_length], ecx
    #BA,#00,#00,#0B,#00,    --   31: mov edx, #B0000
    #29,#CA,                --   36: sub edx, ecx
    #89,#15,#34,#11,#C6,#83,--   38: mov [@bank_limit], edx
    #89,#DA,                --   3E: mov edx, ebx
    #81,#EA,#00,#00,#0A,#00,--   40: sub edx, #A0000
    #C1,#EA,#00,            --   46: shr edx, bank_order (72)
    #B8,#02,#FF,#00,#00,    --   49: mov eax, #FF02
    #89,#15,#92,#11,#C6,#83,--   4E: mov dword ptr [@bank_pos], edx
    #BF,#00,#00,#00,#00,    --   54: mov edi, bank_regs (85)
    #89,#57,#14,            --   59: mov [edi+#14], edx
    #C1,#E2,#00,            --   5C: shl edx, bank_order (94)
    #CD,#31,                --   5F: int #31
    #29,#D3,                --   61: sub ebx, edx
    #89,#DF,                --   63: mov edi, ebx
    #F7,#DB,                --   65: neg ebx
    #81,#E3,#03,#00,#00,#00,--   67: and ebx, #100000003
    #29,#D9,                --   6D: sub ecx, ebx
    #7D,#04,                --   6F: jge line1ok
    #01,#CB,                --   71: add ebx, ecx
    #31,#C9,                --   73: xor ecx, ecx
    #89,#1D,#2E,#11,#C6,#83,--   75: line1ok: mov [@line1], ebx
    #88,#CB,                --   7B: mov bl, cl
    #C1,#E9,#02,            --   7D: shr ecx, 2
    #80,#E3,#03,            --   80: and bl, 3
    #89,#0D,#3D,#11,#C6,#83,--   83: mov [@line2], ecx
    #89,#1D,#44,#11,#C6,#83,--   89: mov [@line3], ebx
    #EB,#00,                --   8F: jmp line_loop
    #B9,#00,#00,#00,#00,    --   91: line_loop: mov ecx, line1 (146)
    #81,#FF,#00,#00,#00,#00,--   96: line_start: cmp edi, dword bank_limit (152)
    #7D,#2E,                --   9C: jge bank_it
    #F3,#A4,                --   9E: rep movsb
    #B9,#00,#00,#00,#00,    --   A0: mov ecx, line2 (161)
    #F3,#A5,                --   A5: rep movsd
    #B9,#00,#00,#00,#00,    --   A7: mov ecx, line3 (168)
    #F3,#A4,                --   AC: rep movsb
    #81,#C6,#00,#00,#00,#00,--   AE: bank_done:add esi, source_add (176)
    #81,#C7,#00,#00,#00,#00,--   B4: add edi, dest_add (182)
    #4D,                    --   BA: dec ebp
    #75,#D4,                --   BB: jnz line_loop
    #BB,#00,#00,#00,#00,    --   BD: mov ebx, saved_ebx (190)
    #8B,#03,                --   C2: mov eax, [ebx]
    #85,#C0,                --   C4: test eax, eax
    #74,#02,                --   C6: jz done
    #FF,#E0,                --   C8: jmp near eax
    #61,                    --   CA: done: popa
    #C3,                    --   CB: ret
    #81,#05,#92,#11,#C6,#83,#00,#00,#00,#00,--   CC: bank_it: add dword ptr [@bank_pos], dword bank_64k (210)
    #89,#FB,                --   D6: mov ebx, edi
    #B9,#00,#00,#00,#00,    --   D8: mov ecx, line_length (217)
    #2B,#1D,#34,#11,#C6,#83,--   DD: sub ebx, [@bank_limit]
    #29,#D9,                --   E3: sub ecx, ebx
    #7E,#02,                --   E5: jle labela
    #F3,#A4,                --   E7: rep movsb
    #B8,#02,#FF,#00,#00,    --   E9: labela: mov eax, #FF02
    #BF,#00,#00,#00,#00,    --   EE: mov edi, bank_regs (239)
    #C7,#47,#14,#00,#00,#00,#00,--   F3: mov dword ptr [edi+#14], bank_pos (246)
    #CD,#31,                --   FA: int #31
    #BF,#00,#00,#0A,#00,    --   FC: mov edi, #A0000
    #01,#CB,                --  101: add ebx, ecx
    #29,#CF,                --  103: sub edi, ecx
    #89,#D9,                --  105: mov ecx, ebx
    #F3,#A4,                --  107: rep movsb
    #EB,#A3})               --  109: jmp bank_done
poke4(bank_copy + 18, bank_copy + 176) -- @source_add
poke4(bank_copy + 31, bank_copy + 182) -- @dest_add
poke4(bank_copy + 37, bank_copy + 190) -- @saved_ebx
poke4(bank_copy + 45, bank_copy + 217) -- @line_length
poke4(bank_copy + 58, bank_copy + 152) -- @bank_limit
poke4(bank_copy + 80, bank_copy + 246) -- @bank_pos
poke4(bank_copy + 119, bank_copy + 146) -- @line1
poke4(bank_copy + 133, bank_copy + 161) -- @line2
poke4(bank_copy + 139, bank_copy + 168) -- @line3
poke4(bank_copy + 206, bank_copy + 246) -- @bank_pos
poke4(bank_copy + 223, bank_copy + 152) -- @bank_limit

map_copy = allocate(111)
poke(map_copy, {
    #8B,#73,#04,            --    0: mov esi, [ebx+4]
    #8B,#7B,#08,            --    3: mov edi, [ebx+8]
    #8B,#4B,#0C,            --    6: mov ecx, [ebx+12]
    #8B,#6B,#10,            --    9: mov ebp, [ebx+16]
    #8B,#43,#14,            --    C: mov eax, [ebx+20]
    #29,#C8,                --    F: sub eax, ecx
    #D1,#E0,                --   11: shl eax, 1
    #A3,#C7,#46,#15,#83,    --   13: mov [@source_add], eax
    #8B,#43,#18,            --   18: mov eax, [ebx+24]
    #29,#C8,                --   1B: sub eax, ecx
    #D1,#E9,                --   1D: shr ecx, 1
    #89,#0D,#AB,#46,#15,#83,--   1F: mov [@width], ecx
    #A3,#CD,#46,#15,#83,    --   25: mov [@dest_add], eax
    #8B,#53,#1C,            --   2A: mov edx, [ebx+28]
    #83,#C3,#20,            --   2D: add ebx, 32
    #89,#1D,#D5,#46,#15,#83,--   30: mov [@saved_ebx], ebx
    #B9,#00,#00,#00,#00,    --   36: label0: mov ecx, width (55)
    #AD,                    --   3B: label1: lodsd
    #89,#C3,                --   3C: mov ebx, eax
    #25,#FF,#FF,#00,#00,    --   3E: and eax, #FFFF
    #C1,#EB,#10,            --   43: shr ebx, 16
    #8A,#04,#02,            --   46: mov al,[eax+edx]
    #8A,#24,#1A,            --   49: mov ah,[ebx+edx]
    #49,                    --   4C: dec ecx
    #66,#AB,                --   4D: stosw
    #75,#EA,                --   4F: jnz label1
    #81,#C6,#00,#00,#00,#00,--   51: add esi, source_add (83)
    #81,#C7,#00,#00,#00,#00,--   57: add edi, dest_add (89)
    #4D,                    --   5D: dec ebp
    #75,#D6,                --   5E: jnz label0
    #BB,#00,#00,#00,#00,    --   60: mov ebx, saved_ebx (97)
    #8B,#03,                --   65: mov eax, [ebx]
    #85,#C0,                --   67: test eax, eax
    #74,#02,                --   69: jz done
    #FF,#E0,                --   6B: jmp near eax
    #61,                    --   6D: done: popa
    #C3})                   --   6E: ret
poke4(map_copy + 20, map_copy + 83) -- @source_add
poke4(map_copy + 33, map_copy + 55) -- @width
poke4(map_copy + 38, map_copy + 89) -- @dest_add
poke4(map_copy + 50, map_copy + 97) -- @saved_ebx


-- ASM copy routines conventions

--    at start, ebx points to the 'kind' value followed by parameters
--    read the parameters from memory, like:
--  "mov eax, [ebx+4] -- first parameter
--  "mov ecx, [ebx+8] -- second parameter
--  "add ebx, 12 "&--    move ebx to the next command, by adding the size of kind & parameters
--  "mov [@saved_ebx], ebx "&--    save ebx if you need to use it for something else
--    ...
--    do your thing
--    ...
--  "mov ebx, saved_ebx "&--    restore ebx if you saved it
--  "mov eax, [ebx] "&--    get the next kind
--  "test eax, eax "&--    test it for zero
--  "jz done "&--      jump if it is
--  "jmp near eax "&--    otherwise, jump to next copy routine
--"done: "&
--  "popa "&   -- pop regs
--  "ret "   -- return
